"""
The :class:`MockAsynctionSocketIO` server is essentially
an :class:`AsynctionSocketIO` server that:

* Periodically emits events containing payloads of fake data,
  through tasks running on the background.
* Listens for all events defined in the given AsyncAPI specification,
  returning fake acknowledgmentds where applicable.
"""
import threading
from functools import partial
from pathlib import Path
from queue import Queue
from random import choice
from typing import Callable
from typing import Mapping
from typing import MutableSequence
from typing import Optional
from typing import Sequence

from faker import Faker
from faker.exceptions import UnsupportedFeature
from flask import Flask
from hypothesis import HealthCheck
from hypothesis import Phase
from hypothesis import Verbosity
from hypothesis import given
from hypothesis import settings
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import sampled_from
from hypothesis_jsonschema import from_schema
from hypothesis_jsonschema._from_schema import STRING_FORMATS

from asynction.server import AsynctionSocketIO
from asynction.types import AsyncApiSpec
from asynction.types import ErrorHandler
from asynction.types import JSONMapping
from asynction.types import JSONSchema
from asynction.types import Message
from asynction.validation import bindings_validator_factory
from asynction.validation import publish_message_validator_factory

CustomFormats = Mapping[str, SearchStrategy[str]]


def make_faker_formats(faker: Faker, sample_size: int) -> CustomFormats:
    custom_formats: CustomFormats = {}
    if sample_size < 1:
        return custom_formats

    for attr in dir(faker):
        if (
            not attr.startswith("_")
            and attr not in Faker.generator_attrs
            and attr not in STRING_FORMATS
        ):
            try:
                provider = getattr(faker, attr)
                if isinstance(provider(), str):
                    custom_formats = {
                        **custom_formats,
                        attr: sampled_from([provider() for _ in range(sample_size)]),
                    }
            except (TypeError, UnsupportedFeature):
                # Skip legacy providers or providers that require extra dependencies
                continue

    return custom_formats


def generate_fake_data_from_schema(
    schema: JSONSchema,
    custom_formats: CustomFormats,
) -> JSONMapping:
    strategy = from_schema(schema, custom_formats=custom_formats)  # type: ignore

    @given(strategy)
    @settings(
        database=None,
        max_examples=30,
        deadline=None,
        verbosity=Verbosity.quiet,
        phases=(Phase.generate,),
        suppress_health_check=HealthCheck.all(),
    )
    def example_generating_inner_function(ex):
        examples.append(ex)

    examples: MutableSequence[JSONMapping] = []
    example_generating_inner_function()
    return choice(examples)


SubscriptionTask = Callable[[], None]


def task_runner(queue: "Queue[SubscriptionTask]") -> None:
    while True:
        task = queue.get()
        task()
        queue.task_done()


def task_scheduler(
    tasks: Sequence[SubscriptionTask],
    queue: "Queue[SubscriptionTask]",
    sleep: Callable[[], None],
) -> None:
    while True:
        for task in tasks:
            queue.put(task)
            sleep()


def _noop_handler(*args, **kwargs) -> None:
    return None


class MockAsynctionSocketIO(AsynctionSocketIO):
    """Inherits the :class:`AsynctionSocketIO` class."""

    def __init__(
        self,
        spec: AsyncApiSpec,
        validation: bool,
        app: Optional[Flask],
        custom_formats_sample_size: int,
        **kwargs,
    ):
        """This is a private constructor.
        Use the :meth:`MockAsynctionSocketIO.from_spec` factory instead.
        """
        super().__init__(spec, validation=validation, app=app, **kwargs)
        self.faker = Faker()
        self.custom_formats = make_faker_formats(self.faker, custom_formats_sample_size)
        self._subscription_tasks: Sequence[SubscriptionTask] = []

    @classmethod
    def from_spec(
        cls,
        spec_path: Path,
        validation: bool = True,
        server_name: Optional[str] = None,
        default_error_handler: Optional[ErrorHandler] = None,
        app: Optional[Flask] = None,
        custom_formats_sample_size: int = 20,
        **kwargs,
    ) -> "MockAsynctionSocketIO":
        """Create a Flask-SocketIO mock server given an AsyncAPI spec.
        The server emits events containing payloads of fake data in regular intervals,
        through background subscription tasks.
        It also listens for events as per the spec definitions
        and returns mock aknowledgements where applicable.
        All event and acknowledgment payloads adhere to the schemata defined
        within the AsyncAPI spec.

        In addition to the args and kwargs of :meth:`AsynctionSocketIO.from_spec`,
        this factory method accepts some extra keyword arguments:

        * ``custom_formats_sample_size``

        :param spec_path: The path where the AsyncAPI YAML specification is located.
        :param validation: When set to ``False``, message payloads, channel
                           bindings and ack callbacks are NOT validated.
                           Defaults to ``True``.
        :param server_name: The server to pick from the AsyncAPI ``servers`` object.
                            The server object is then used to configure
                            the path ``kwarg`` of the SocketIO server.
        :param default_error_handler: The error handler that handles any namespace
                                      without an explicit error handler.
                                      Equivelant of ``@socketio.on_error_default``
        :param app: The flask application instance. Defaults to ``None``.
        :param subscription_task_interval: How often (in seconds) a subscription task
                                           (thread that emits an event to
                                           a connected client) is scheduled.
                                           Defaults to ``1``.
        :param max_worker_number: The maximum number of workers to be started for the
                                  purposes of executing background subscription tasks.
                                  Defaults to ``8``.
        :param custom_formats_sample_size: The ammout of the Faker provider samples
                                           to be used for each custom string format.
                                           Hypotheses uses these samples to generate
                                           fake data. Set to ``0`` if custom formats
                                           are not needed.
                                           Defaults to ``20``.
        :param kwargs: Flask-SocketIO, Socket.IO and Engine.IO server options.

        :returns: A Flask-SocketIO mock server, emitting events with fake data in
                  regular intervals.
                  The server also has mock event and error handlers registered.

        Example::

            mock_asio = MockAsynctionSocketIO.from_spec(
                spec_path="./docs/asyncapi.yaml",
                app=flask_app,
                # any other kwarg that the flask_socketio.SocketIO constructor accepts
            )

        """
        return super().from_spec(
            spec_path,
            validation=validation,
            server_name=server_name,
            default_error_handler=default_error_handler,
            app=app,
            custom_formats_sample_size=custom_formats_sample_size,
            **kwargs,
        )

    def _register_handlers(
        self, default_error_handler: Optional[ErrorHandler] = None
    ) -> None:
        for namespace, channel in self.spec.channels.items():
            if channel.publish is not None:
                for message in channel.publish.message.oneOf:
                    handler = self.make_publish_handler(message)

                    if self.validation:
                        with_payload_validation = publish_message_validator_factory(
                            message=message
                        )
                        handler = with_payload_validation(handler)

                    self.on_event(message.name, handler, namespace)

            if channel.subscribe is not None:
                self._subscription_tasks = [
                    *self._subscription_tasks,
                    *[
                        self.make_subscription_task(message, namespace)
                        for message in channel.subscribe.message.oneOf
                    ],
                ]

            connect_handler = _noop_handler

            if self.validation:
                with_bindings_validation = bindings_validator_factory(channel.bindings)
                connect_handler = with_bindings_validation(connect_handler)

            self.on_event("connect", connect_handler, namespace)

        if default_error_handler is not None:
            self.on_error_default(default_error_handler)

    def make_subscription_task(
        self, message: Message, namespace: str
    ) -> SubscriptionTask:
        def task() -> None:
            self.emit(
                message.name,
                generate_fake_data_from_schema(
                    message.payload or {"type": "null"}, self.custom_formats
                ),
                namespace=namespace,
                callback=message.x_ack and _noop_handler,
            )

        return task

    def make_publish_handler(self, message: Message) -> Callable:
        if message.x_ack is not None:

            def handler(*args, **kwargs):
                return generate_fake_data_from_schema(
                    message.x_ack.args, self.custom_formats
                )

            return handler

        return _noop_handler

    def start_background_task(
        self, target: Callable, *args, **kwargs
    ) -> threading.Thread:

        # The tasks created in the :meth:`MockAsynctionSocketIO.run` method below
        # (both runner and scheduler) MUST be daemonic.
        # However, python-engineio does not support daemonic background tasks,
        # unless the chosen async mode defaults to some daemon-like behaviour.
        # Native threads have daemon set to False by default, which is rather
        # inconvinient for this use case.
        # See the relevant issue:
        # https://github.com/miguelgrinberg/python-engineio/issues/244
        #
        # The current method is a hack that accounts for the threading scenario,
        # to ensure that native threads are started as daemons.

        if self.async_mode == "threading":
            th = threading.Thread(target=target, args=args, kwargs=kwargs, daemon=True)
            th.start()
            return th

        return super().start_background_task(target, *args, **kwargs)

    def run(
        self,
        app: Flask,
        host: Optional[str] = None,
        port: Optional[int] = None,
        subscription_task_interval: float = 1.0,
        max_worker_number: int = 8,
        **kwargs,
    ) -> None:
        """
        Run the mock Asynction SocketIO web server.

        In addition to the args and kwargs of :meth:`flask_socketio.SocketIO.run`,
        this method accepts some extra keyword arguments:

        * ``subscription_task_interval``
        * ``max_worker_number``

        :param app: The flask application instance.
        :param host: The hostname or IP address for the server to listen on.
                     Defaults to ``127.0.0.1``.
        :param port: The port number for the server to listen on. Defaults to
                     ``5000``.
        :param subscription_task_interval: How often (in seconds) a subscription task
                                           (thread that emits an event to
                                           a connected client) is scheduled.
                                           Defaults to ``1.0``.
        :param max_worker_number: The maximum number of workers to be started for the
                                  purposes of executing background subscription tasks.
                                  Defaults to ``8``.
        :param kwargs: Additional web server options that are propagated to
                       :meth:`flask_socketio.SocketIO.run`. The web server options
                       are specific to the server used in each of the supported
                       async modes. Refer to the Flask-SocketIO docs for details.
        """
        queue: "Queue[SubscriptionTask]" = self.server.eio.create_queue()

        for _ in range(min(max_worker_number, len(self._subscription_tasks))):
            _ = self.start_background_task(task_runner, queue=queue)

        _ = self.start_background_task(
            task_scheduler,
            tasks=self._subscription_tasks,
            queue=queue,
            sleep=partial(self.sleep, subscription_task_interval),
        )

        return super().run(app, host=host, port=port, **kwargs)



from functools import partial
from functools import wraps
from typing import Callable
from typing import Optional
from typing import Sequence
from typing import Type

import jsonschema
from flask import Request
from flask import request as current_flask_request

from asynction.exceptions import BindingsValidationException
from asynction.exceptions import MessageAckValidationException
from asynction.exceptions import PayloadValidationException
from asynction.types import ChannelBindings
from asynction.types import JSONMapping
from asynction.types import JSONSchema
from asynction.types import Message
from asynction.types import MessageAck


def jsonschema_validate_with_error_handling(
    instance: JSONMapping, schema: JSONSchema, exc_type: Type[Exception]
) -> None:
    try:
        jsonschema.validate(instance, schema)
    except jsonschema.ValidationError as e:
        raise exc_type(str(e))


jsonschema_validate_payload = partial(
    jsonschema_validate_with_error_handling,
    exc_type=PayloadValidationException,
)

jsonschema_validate_bindings = partial(
    jsonschema_validate_with_error_handling, exc_type=BindingsValidationException
)

jsonschema_validate_ack = partial(
    jsonschema_validate_with_error_handling, exc_type=MessageAckValidationException
)


def validate_payload(args: Sequence, schema: Optional[JSONSchema]) -> None:
    if schema is None:
        if args:
            raise PayloadValidationException(
                "Handler args provided for message that has no payload defined."
            )
        # Validation succeeded since there is no message payload specified
        # and no args have been provided.
        return

    schema_type = schema["type"]
    if schema_type == "array":
        jsonschema_validate_payload(args, schema)
    else:
        if len(args) > 1:
            raise PayloadValidationException(
                "Multiple handler arguments provided, "
                f"although schema type is: {schema_type}"
            )

        jsonschema_validate_payload(args[0], schema)


def validate_ack_args(args: Sequence, message_ack_spec: Optional[MessageAck]) -> None:
    if message_ack_spec is None:
        if args:
            raise MessageAckValidationException(
                "Callback args provided for message that has no x-ack schema specified."
            )
        # Validation succeeded since there is no message ack specified
        # and no args have been provided.
        return

    schema_type = message_ack_spec.args["type"]
    if schema_type == "array":
        jsonschema_validate_ack(args, message_ack_spec.args)
    else:
        if len(args) > 1:
            raise MessageAckValidationException(
                "Multiple callback arguments provided, "
                f"although x-ack schema type is: {schema_type}"
            )

        jsonschema_validate_ack(args[0], message_ack_spec.args)


def publish_message_validator_factory(message: Message) -> Callable:
    """Constructs a validating wrapper for any incoming (`publish`) message handler"""

    def decorator(handler: Callable):
        @wraps(handler)
        def handler_with_validation(*args):
            validate_payload(args, message.payload)

            ack = handler(*args)

            if ack is not None and message.x_ack is not None:
                jsonschema_validate_ack(ack, message.x_ack.args)

            # TODO: Should the ack be transmitted if there is no x-ack definition?
            return ack

        return handler_with_validation

    return decorator


def callback_validator_factory(message: Message) -> Callable:
    def decorator(callback: Callable):
        @wraps(callback)
        def callback_with_validation(*args):
            # the callback should only be called with positional arguments
            validate_ack_args(args, message.x_ack)
            return callback(*args)

        return callback_with_validation

    return decorator


def validate_request_bindings(
    request: Request, bindings: Optional[ChannelBindings]
) -> None:
    if bindings is None:
        # No validation needed since there are no bindings defined
        return

    if bindings.ws.method is not None and request.method != bindings.ws.method:
        raise BindingsValidationException(
            "Binding validation failed. "
            f"Request method {request.method} does not match {bindings.ws.method}."
        )

    if bindings.ws.headers is not None:
        jsonschema_validate_bindings(
            {k.lower(): v for k, v in request.headers.items()}, bindings.ws.headers
        )

    if bindings.ws.query is not None:
        jsonschema_validate_bindings(request.args.to_dict(), bindings.ws.query)


def bindings_validator_factory(bindings: Optional[ChannelBindings]) -> Callable:
    def decorator(handler: Callable):
        @wraps(handler)
        def handler_with_validation(*args):
            validate_request_bindings(current_flask_request, bindings)
            return handler(*args)

        return handler_with_validation

    return decorator



"""Asynction's exceptions to be caught via Flask-SocketIO error handlers."""


class AsynctionException(RuntimeError):
    """The base class for all asynction runtime exceptions."""

    pass


class ValidationException(AsynctionException):
    """The base class for all asynction validation exceptions."""

    pass


class PayloadValidationException(ValidationException):
    """
    Raised when the payload of an incoming or outgoing message
    fails the schema validation.
    """

    pass


class BindingsValidationException(ValidationException):
    """
    Raised when the HTTP bindings of an incoming connection
    fail the schema validation.
    """

    pass


class MessageAckValidationException(ValidationException):
    """
    Raised when the input arguments passed to the ack callback
    do not adhere to the ``x-ack`` args schema.
    """

    pass



from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Any
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Type

from svarog import forge
from svarog import register_forge
from svarog.types import Forge

JSONMappingValue = Any
JSONMapping = Mapping[str, JSONMappingValue]
JSONSchema = JSONMapping

GLOBAL_NAMESPACE = "/"


@dataclass
class MessageAck:
    """The specification of a message acknowledgement"""

    args: JSONSchema


@dataclass
class Message:
    """
    https://www.asyncapi.com/docs/specifications/2.0.0#messageObject

    The above message object is extended as follows:
    * `x-handler`: Allows the coupling of the message specification to
    an event handler (which is a python callable). It SHOULD only be used
    for messages under a `publish` operation. Deserialized to `x_handler`.
    * `x-ack`: The specification of the acknowledgement packet that the message receiver
    transmits to the message sender. The acknowledgement args are passed as an input
    to the callback of the `emit`/`send` function. Deserialized to `x_ack`.

    The extentions are implemented as per:
    https://www.asyncapi.com/docs/specifications/2.0.0#specificationExtensions
    """

    name: str
    payload: Optional[JSONSchema] = None
    x_handler: Optional[str] = None
    x_ack: Optional[MessageAck] = None

    @staticmethod
    def forge(type_: Type["Message"], data: JSONMapping, forge: Forge) -> "Message":
        return type_(
            name=forge(type_.__annotations__["name"], data["name"]),
            payload=forge(type_.__annotations__["payload"], data.get("payload")),
            x_handler=forge(type_.__annotations__["x_handler"], data.get("x-handler")),
            x_ack=forge(type_.__annotations__["x_ack"], data.get("x-ack")),
        )


register_forge(Message, Message.forge)


@dataclass
class OneOfMessages:
    """Using `oneOf` to specify multiple messages per operation"""

    oneOf: Sequence[Message]

    @staticmethod
    def forge(
        type_: Type["OneOfMessages"], data: JSONMapping, forge: Forge
    ) -> "OneOfMessages":
        if "oneOf" in data:
            return type_(
                oneOf=forge(type_.__annotations__["oneOf"], data["oneOf"]),
            )

        return type_(oneOf=[forge(Message, data)])

    def with_name(self, name: str) -> Optional[Message]:
        for message in self.oneOf:
            if message.name == name:
                return message

        return None


register_forge(OneOfMessages, OneOfMessages.forge)


@dataclass
class Operation:
    """https://www.asyncapi.com/docs/specifications/2.0.0#operationObject"""

    message: OneOfMessages


@dataclass
class WebSocketsChannelBindings:
    """
    https://github.com/asyncapi/bindings/tree/master/websockets#channel-binding-object
    """

    method: Optional[str] = None
    query: Optional[JSONSchema] = None
    headers: Optional[JSONSchema] = None  # TODO: Convert header properties to lowercase
    bindingVersion: str = "latest"


@dataclass
class ChannelBindings:
    """https://www.asyncapi.com/docs/specifications/2.0.0#channelBindingsObject"""

    ws: WebSocketsChannelBindings


@dataclass
class ChannelHandlers:
    connect: Optional[str] = None
    disconnect: Optional[str] = None
    error: Optional[str] = None


@dataclass
class Channel:
    """
    https://www.asyncapi.com/docs/specifications/2.0.0#channelItemObject

    The above channel item object is extended to
    support default namespace handlers as per:
    https://www.asyncapi.com/docs/specifications/2.0.0#specificationExtensions

    The `x_handlers` field is serialized as `x-handlers`.
    """

    subscribe: Optional[Operation] = None
    publish: Optional[Operation] = None
    bindings: Optional[ChannelBindings] = None
    x_handlers: Optional[ChannelHandlers] = None

    def __post_init__(self):
        if self.publish is not None:
            for message in self.publish.message.oneOf:
                if message.x_handler is None:
                    raise ValueError(
                        f"Message {message.name} is missing the x-handler attribute.\n"
                        "Every message under a publish operation "
                        "should have a handler defined."
                    )

    @staticmethod
    def forge(type_: Type["Channel"], data: JSONMapping, forge: Forge) -> "Channel":
        return type_(
            subscribe=forge(type_.__annotations__["subscribe"], data.get("subscribe")),
            publish=forge(type_.__annotations__["publish"], data.get("publish")),
            bindings=forge(type_.__annotations__["bindings"], data.get("bindings")),
            x_handlers=forge(
                type_.__annotations__["x_handlers"], data.get("x-handlers")
            ),
        )


register_forge(Channel, Channel.forge)


class ServerProtocol(Enum):
    HTTP = "http"
    HTTPS = "https"
    WS = "ws"
    WSS = "wss"


@dataclass
class Server:
    """https://www.asyncapi.com/docs/specifications/2.0.0#serverObject"""

    url: str
    protocol: ServerProtocol


@dataclass
class AsyncApiSpec:
    """https://www.asyncapi.com/docs/specifications/2.0.0#A2SObject"""

    channels: Mapping[str, Channel]
    servers: Mapping[str, Server] = field(default_factory=dict)

    @staticmethod
    def from_dict(data: JSONMapping) -> "AsyncApiSpec":
        return forge(AsyncApiSpec, data)


ErrorHandler = Callable[[Exception], None]



# flake8: noqa: F401

__all__ = [
    "AsynctionSocketIO",
    "AsynctionException",
    "ValidationException",
    "PayloadValidationException",
    "BindingsValidationException",
    "MessageAckValidationException",
]

from asynction.exceptions import *
from asynction.server import AsynctionSocketIO

try:
    from asynction.mock_server import MockAsynctionSocketIO

    __all__ = [*__all__, "MockAsynctionSocketIO"]
except ImportError:
    # Mock support may not be available if the mock
    # extra requirements have not been installed
    pass



import eventlet

eventlet.monkey_patch()  # noreorder

import argparse
from functools import singledispatch
from pathlib import Path
from typing import Optional

from flask import Flask
from importlib_metadata import PackageNotFoundError
from importlib_metadata import version

import asynction


def get_version() -> str:
    try:
        return version(asynction.__name__)
    except PackageNotFoundError:
        return "not-found"


class AsynctionNamespace(argparse.Namespace):
    command: str
    spec: Path
    server: Optional[str]
    without_validation: bool


class RunNamespace(AsynctionNamespace):
    host: str
    port: int
    debugger: bool


class MockNamespace(AsynctionNamespace):
    mock_command: str
    custom_formats_sample_size: int


class RunMockNamespace(MockNamespace, RunNamespace):
    subscription_task_interval: float
    max_workers: int


def build_parser() -> argparse.ArgumentParser:
    formatter_class = argparse.ArgumentDefaultsHelpFormatter

    parser = argparse.ArgumentParser(
        asynction.__name__, formatter_class=formatter_class
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {get_version()}",
    )
    parser.add_argument(
        "--spec",
        "-s",
        type=Path,
        default=Path.cwd().joinpath("asyncapi.yaml"),
        help="Path to the AsyncAPI specification YAML file",
    )
    parser.add_argument(
        "--without-validation",
        default=False,
        action="store_true",
        help="Disable Asynction's default validation",
    )
    parser.add_argument(
        "--server",
        required=False,
        help="The server to pick from the AsyncAPI servers object",
    )

    subparsers = parser.add_subparsers(title="Sub-commands", dest="command")
    mock_parser = subparsers.add_parser(
        "mock",
        formatter_class=formatter_class,
        help="Utilities for the mock Asynction SocketIO server",
    )
    mock_parser.add_argument(
        "--custom-formats-sample-size",
        type=int,
        default=20,
        help="The ammout of the Faker provider samples to be used "
        "for each custom string format. "
        "Hypotheses uses these samples to generate fake data. "
        "Set to 0 if custom formats are not needed.",
    )

    mock_subparsers = mock_parser.add_subparsers(
        title="Mock sub-commands", dest="mock_command"
    )
    run_mock_parser = mock_subparsers.add_parser(
        "run",
        formatter_class=formatter_class,
        help="Run a mock Asynction SocketIO server",
    )
    run_mock_parser.add_argument(
        "--host", default="0.0.0.0", help="The interface to bind to"
    )
    run_mock_parser.add_argument(
        "--port", "-p", type=int, default=5000, help="The port to bind to"
    )
    run_mock_parser.add_argument(
        "--subscription-task-interval",
        type=float,
        default=1.0,
        help="How often (in seconds) a subscription task "
        "(thread that emits an event to a connected client) is scheduled.",
    )
    run_mock_parser.add_argument(
        "--max-workers",
        type=int,
        default=8,
        help="The maximum number of workers to be started for the purposes "
        "of executing background subscription tasks.",
    )

    run_mock_parser.add_argument(
        "--debugger",
        "-d",
        action="store_true",
        help="Enable the debugger",
    )

    return parser


@singledispatch
def command(args: AsynctionNamespace, parser: argparse.ArgumentParser) -> None:
    if args.command == "mock":
        return command(MockNamespace(**vars(args)), parser)

    parser.print_help()


@command.register
def _mock_command(args: MockNamespace, parser: argparse.ArgumentParser) -> None:
    if args.mock_command == "run":
        return command(RunMockNamespace(**vars(args)), parser)

    parser.parse_args([args.command, "--help"])


@command.register
def _run_mock_command(args: RunMockNamespace, parser: argparse.ArgumentParser) -> None:
    flask_app = Flask(__name__)

    mock_asio = asynction.MockAsynctionSocketIO.from_spec(
        spec_path=args.spec,
        validation=not args.without_validation,
        server_name=args.server,
        logger=args.debugger,
        app=flask_app,
        cors_allowed_origins="*",
        custom_formats_sample_size=args.custom_formats_sample_size,
    )

    mock_asio.run(
        app=flask_app,
        debug=args.debugger,
        log_output=args.debugger,
        host=args.host,
        port=args.port,
        subscription_task_interval=args.subscription_task_interval,
        max_worker_number=args.max_workers,
    )


def main() -> None:
    parser = build_parser()
    args = parser.parse_args(namespace=AsynctionNamespace())
    return command(args, parser)



#!/usr/bin/env python

import os
from pathlib import Path
from typing import Mapping
from typing import Sequence

from setuptools import find_packages
from setuptools import setup

with open("README.md") as readme_file:
    readme = readme_file.read()


with open("requirements.txt") as requiremets_file:
    requirements = requiremets_file.read().split()


def parse_requirements(file_path: Path) -> Sequence[str]:
    reqs = []
    with file_path.open() as f:
        raw_reqs = f.read().split(os.linesep)
        for raw_req in raw_reqs:
            if raw_req:
                stripped_raw_req = raw_req.strip()

                expanded_reqs = (
                    [stripped_raw_req]
                    if not stripped_raw_req.startswith("-r ")
                    else parse_requirements(
                        Path.cwd().joinpath(stripped_raw_req[len("-r ") :].strip())
                    )
                )

                reqs = [*reqs, *expanded_reqs]

    return reqs


def make_extra_requirements() -> Mapping[str, str]:
    extra_requirements = {}
    for req_file_path in Path(__file__).parent.glob("requirements-*.txt"):
        if req_file_path.name not in ["requirements-dev.txt", "requirements-test.txt"]:
            extra_req_name = req_file_path.name[len("requirements-") : -len(".txt")]
            extra_requirements = {
                **extra_requirements,
                extra_req_name: parse_requirements(req_file_path),
            }

    return extra_requirements


version = os.environ["PKG_VERSION"]


setup(
    author="Dimitrios Dedoussis",
    author_email="dimitrios@dedouss.is",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Framework :: Flask",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    description="SocketIO framework driven by the AsyncAPI specification. Built on top of Flask-SocketIO. Inspired by Connexion.",
    install_requires=requirements,
    extras_require=make_extra_requirements(),
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    package_data={
        "asynction": ["py.typed"],
    },
    keywords=" ".join(
        [
            "asyncapi",
            "websockets",
            "socketio",
            "socket.io",
            "api",
            "oauth",
            "flask",
            "microservice",
            "framework",
            "specification",
            "flask-socketio",
            "connexion",
            "mock",
        ]
    ),
    name="asynction",
    packages=find_packages(include=["asynction", "asynction.*"]),
    entry_points={"console_scripts": ["asynction = asynction.__main__:main"]},
    url="https://github.com/dedoussis/asynction",
    version=version,
    zip_safe=False,
)



"""
The :class:`AsynctionSocketIO` server is essentially a ``flask_socketio.SocketIO``
server with an additional factory classmethod.
"""

from functools import singledispatch
from importlib import import_module
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Optional
from typing import Sequence
from urllib.parse import urlparse

import jsonschema
import yaml
from flask import Flask
from flask_socketio import SocketIO

from asynction.exceptions import ValidationException
from asynction.types import GLOBAL_NAMESPACE
from asynction.types import AsyncApiSpec
from asynction.types import ChannelBindings
from asynction.types import ChannelHandlers
from asynction.types import ErrorHandler
from asynction.types import JSONMapping
from asynction.validation import bindings_validator_factory
from asynction.validation import callback_validator_factory
from asynction.validation import publish_message_validator_factory
from asynction.validation import validate_payload


@singledispatch
def deep_resolve(unresovled: Any, resolver: jsonschema.RefResolver) -> Any:
    return unresovled


@deep_resolve.register(dict)
def _deep_resolve_mapping(
    unresolved: JSONMapping, resolver: jsonschema.RefResolver
) -> JSONMapping:
    return {
        k: deep_resolve(resolver.resolve(v["$ref"])[-1] if "$ref" in v else v, resolver)
        for k, v in unresolved.items()
    }


@deep_resolve.register(list)
@deep_resolve.register(tuple)
@deep_resolve.register(set)
def _deep_resolve_sequence(
    unresolved: Sequence, resolver: jsonschema.RefResolver
) -> Sequence:
    return unresolved.__class__(  # type: ignore
        [
            deep_resolve(
                resolver.resolve(item["$ref"])[-1] if "$ref" in item else item, resolver
            )
            for item in unresolved
        ]
    )


def resolve_references(raw_spec: JSONMapping) -> JSONMapping:
    resolver = jsonschema.RefResolver.from_schema(raw_spec)
    return deep_resolve(raw_spec, resolver)


def load_spec(spec_path: Path) -> AsyncApiSpec:
    with open(spec_path) as f:
        serialized = f.read()
        raw = yaml.safe_load(serialized)

    raw_resolved = resolve_references(raw_spec=raw)

    return AsyncApiSpec.from_dict(raw_resolved)


def load_handler(handler_id: str) -> Callable:
    *module_path_elements, object_name = handler_id.split(".")
    module = import_module(".".join(module_path_elements))

    return getattr(module, object_name)


class AsynctionSocketIO(SocketIO):
    """Inherits the :class:`flask_socketio.SocketIO` class."""

    def __init__(
        self,
        spec: AsyncApiSpec,
        validation: bool = True,
        app: Optional[Flask] = None,
        **kwargs,
    ):
        """This is a private constructor.
        Use the :meth:`AsynctionSocketIO.from_spec` factory instead.
        """
        super().__init__(app=app, **kwargs)
        self.spec = spec
        self.validation = validation

    @classmethod
    def from_spec(
        cls,
        spec_path: Path,
        validation: bool = True,
        server_name: Optional[str] = None,
        default_error_handler: Optional[ErrorHandler] = None,
        app: Optional[Flask] = None,
        **kwargs,
    ) -> SocketIO:
        """Create a Flask-SocketIO server from an AsyncAPI spec.
        This is the single entrypoint to the Asynction server API.

        :param spec_path: The path where the AsyncAPI YAML specification is located.
        :param validation: When set to ``False``, message payloads, channel
                           bindings and ack callbacks are NOT validated.
                           Defaults to ``True``.
        :param server_name: The server to pick from the AsyncAPI ``servers`` object.
                            The server object is then used to configure
                            the path ``kwarg`` of the SocketIO server.
        :param default_error_handler: The error handler that handles any namespace
                                      without an explicit error handler.
                                      Equivelant of ``@socketio.on_error_default``
        :param app: The flask application instance. Defaults to ``None``.
        :param kwargs: Flask-SocketIO, Socket.IO and Engine.IO server options.

        :returns: A Flask-SocketIO server.
                  The server has all the event and error handlers registered.

        Example::

            asio = AsynctionSocketIO.from_spec(
                spec_path="./docs/asyncapi.yaml",
                app=flask_app,
                message_queue="redis://localhost:6379",
                # any other kwarg that the flask_socketio.SocketIO constructor accepts
            )

        """
        spec = load_spec(spec_path=spec_path)

        if (
            server_name is not None
            and kwargs.get("path") is None
            and kwargs.get("resource") is None
        ):
            server = spec.servers.get(server_name)
            if server is None:
                raise ValueError(f"Server {server_name} is not defined in the spec.")

            url_parse_result = urlparse(
                url=f"//{server.url}", scheme=server.protocol.value
            )

            if url_parse_result.path:
                kwargs["path"] = url_parse_result.path

        asio = cls(spec, validation, app, **kwargs)
        asio._register_handlers(default_error_handler)
        return asio

    def _register_namespace_handlers(
        self,
        namespace: str,
        channel_handlers: ChannelHandlers,
        channel_bindings: Optional[ChannelBindings],
    ) -> None:
        if channel_handlers.connect is not None:
            handler = load_handler(channel_handlers.connect)

            if self.validation:
                with_bindings_validation = bindings_validator_factory(channel_bindings)
                handler = with_bindings_validation(handler)

            self.on_event("connect", handler, namespace)

        if channel_handlers.disconnect is not None:
            handler = load_handler(channel_handlers.disconnect)
            self.on_event("disconnect", handler, namespace)

        if channel_handlers.error is not None:
            handler = load_handler(channel_handlers.error)
            self.on_error(namespace)(handler)

    def _register_handlers(
        self, default_error_handler: Optional[ErrorHandler] = None
    ) -> None:
        for namespace, channel in self.spec.channels.items():
            if channel.publish is not None:
                for message in channel.publish.message.oneOf:
                    assert message.x_handler is not None
                    handler = load_handler(message.x_handler)

                    if self.validation:
                        with_payload_validation = publish_message_validator_factory(
                            message=message
                        )
                        handler = with_payload_validation(handler)

                    self.on_event(message.name, handler, namespace)

            if channel.x_handlers is not None:
                self._register_namespace_handlers(
                    namespace, channel.x_handlers, channel.bindings
                )

        if default_error_handler is not None:
            self.on_error_default(default_error_handler)

    def emit(self, event: str, *args, **kwargs) -> None:
        if self.validation:
            namespace = kwargs.get("namespace", GLOBAL_NAMESPACE)
            channel = self.spec.channels.get(namespace)

            if channel is None:
                raise ValidationException(
                    f"Failed to emit because the {namespace} "
                    "namespace is not defined in the API spec."
                )

            if channel.subscribe is None:
                raise ValidationException(
                    f"Failed to emit because the {namespace} namespace "
                    "does not have any subscribe operations defined."
                )

            message = channel.subscribe.message.with_name(event)
            if message is None:
                raise ValidationException(
                    f"Event {event} is not registered under the {namespace} namespace"
                )

            validate_payload(args, message.payload)

            callback = kwargs.get("callback")
            if callback is not None:
                with_validation = callback_validator_factory(message=message)
                kwargs["callback"] = with_validation(callback)

        return super().emit(event, *args, **kwargs)
